@@ -2,10 +2,17 @@ import decimal
 import itertools
 import math
 import operator
-from collections.abc import Iterable
+import sys
 from decimal import Decimal, localcontext
-from typing import Any, Optional, SupportsFloat, SupportsIndex, Type
-from typing import TypeVar, Union, get_args, overload
+from typing import Any, Optional, SupportsFloat, SupportsIndex, TypeVar
+from typing import Union, get_args, overload
+
+if sys.version_info < (3, 9):
+    from typing import Dict, Iterable, List, Tuple, Type
+else:
+    from builtins import dict as Dict, list as List, tuple as Tuple
+    from builtins import type as Type
+    from collections.abc import Iterable
 
 from .fpu_rounding import *
 from .interval import NOT_REAL, Interval, interval
@@ -14,10 +21,10 @@ from .typing import RealLike, SupportsRichFloat
 
 NOT_REAL = "could not interpret {} as an interval"
 
-e = interval[math.e:math.nextafter(math.e, math.inf)]
+e = interval[math.e:nextafter(math.e, math.inf)]
 inf = interval[math.inf:]
 
-_PI = interval[math.pi:math.nextafter(math.pi, math.inf)]
+_PI = interval[math.pi:nextafter(math.pi, math.inf)]
 _BIG_PI = Decimal(
     "3."
     "14159265358979323846264338327950288419716939937510582097494459230"
@@ -90,7 +97,7 @@ class PiMultiple(Interval):
         return self.coefficients * _PI
 
     @classmethod
-    def __dist__(cls: Type[Self], p: list[Interval], q: list[Interval]) -> Self:
+    def __dist__(cls: Type[Self], p: List[Interval], q: List[Interval]) -> Self:
         if not all(
             type(x).__dist__.__func__ is cls.__dist__.__func__
             for pq in (p, q)
@@ -121,7 +128,7 @@ class PiMultiple(Interval):
             return format(self.coefficients, specifier) + " * pi"
 
     @classmethod
-    def __fsum__(cls: Type[Self], intervals: list[Interval]) -> Self:
+    def __fsum__(cls: Type[Self], intervals: List[Interval]) -> Self:
         if not all(
             type(x).__fsum__.__func__ is cls.__fsum__.__func__
             for x in intervals
@@ -129,7 +136,7 @@ class PiMultiple(Interval):
             return NotImplemented
         return pi * fsum(x.coefficients for x in intervals)
 
-    def __getitem__(self: Self, args: Union[slice, tuple[slice, ...]]) -> Interval:
+    def __getitem__(self: Self, args: Union[slice, Tuple[slice, ...]]) -> Interval:
         return self.__as_interval__()[args]
 
     def __mul__(self: Self, other: Union[Interval, float]) -> Interval:
@@ -330,7 +337,7 @@ def sin_precise(x: Decimal) -> Decimal:
             s += sign * num / fact
         return s
 
-def cos_sin_precise(x: Union[Decimal, SupportsIndex, float]) -> tuple[Decimal, Decimal]:
+def cos_sin_precise(x: Union[Decimal, SupportsIndex, float]) -> Tuple[Decimal, Decimal]:
     with localcontext() as ctx:
         if isinstance(x, SupportsIndex):
             x = operator.index(x)
@@ -373,14 +380,14 @@ def acos(x: Union[Interval, RealLike]) -> Interval:
 def acos_down(x: float) -> float:
     y = math.acos(x)
     if cos_sin_precise(y)[0] < x:
-        return math.nextafter(y, 0.0)
+        return nextafter(y, 0.0)
     else:
         return y
 
 def acos_up(x: float) -> float:
     y = math.acos(x)
     if cos_sin_precise(y)[0] > x:
-        return math.nextafter(y, 4.0)
+        return nextafter(y, 4.0)
     else:
         return y
 
@@ -401,7 +408,7 @@ def acosh_down(x: float) -> float:
     y = math.acosh(x)
     d = Decimal(y)
     if (d.exp() + (-d).exp()) / 2 > x:
-        return math.nextafter(y, 0.0)
+        return nextafter(y, 0.0)
     else:
         return y
 
@@ -409,7 +416,7 @@ def acosh_up(x: float) -> float:
     y = math.acosh(x)
     d = Decimal(y)
     if (d.exp() + (-d).exp()) / 2 < x:
-        return math.nextafter(y, math.inf)
+        return nextafter(y, math.inf)
     else:
         return y
 
@@ -440,7 +447,7 @@ def asin(x: Union[Interval, RealLike]) -> Interval:
 def asin_down(x: float) -> float:
     y = math.asin(x)
     if cos_sin_precise(y)[1] > x:
-        return math.nextafter(y, -math.inf)
+        return nextafter(y, -math.inf)
     else:
         return y
 
@@ -448,7 +455,7 @@ def asin_up(x: float) -> float:
     y = math.asin(x)
     d = Decimal(y)
     if cos_sin_precise(y)[1] < x:
-        return math.nextafter(y, math.inf)
+        return nextafter(y, math.inf)
     else:
         return y
 
@@ -469,7 +476,7 @@ def asinh_down(x: float) -> float:
     y = math.asinh(x)
     d = Decimal(y)
     if (d.exp() - (-d).exp()) / 2 > x:
-        return math.nextafter(y, -math.inf)
+        return nextafter(y, -math.inf)
     else:
         return y
 
@@ -477,7 +484,7 @@ def asinh_up(x: float) -> float:
     y = math.asinh(x)
     d = Decimal(y)
     if (d.exp() - (-d).exp()) / 2 < x:
-        return math.nextafter(y, math.inf)
+        return nextafter(y, math.inf)
     else:
         return y
 
@@ -525,7 +532,7 @@ def atan_down(x: float) -> float:
     y = math.atan(x)
     c, s = cos_sin_precise(y)
     if s / c > x:
-        return math.nextafter(y, -math.inf)
+        return nextafter(y, -math.inf)
     else:
         return y
 
@@ -533,7 +540,7 @@ def atan_up(x: float) -> float:
     y = math.atan(x)
     c, s = cos_sin_precise(y)
     if s / c < x:
-        return math.nextafter(y, math.inf)
+        return nextafter(y, math.inf)
     else:
         return y
 
@@ -642,7 +649,7 @@ def atan2_down(y: float, x: float) -> float:
     z = math.atan2(y, x)
     c, s = cos_sin_precise(z)
     if s / c > Decimal(y) / Decimal(x):
-        return math.nextafter(z, -math.inf)
+        return nextafter(z, -math.inf)
     else:
         return z
 
@@ -650,7 +657,7 @@ def atan2_up(y: float, x: float) -> float:
     z = math.atan2(y, x)
     c, s = cos_sin_precise(z)
     if s / c < Decimal(y) / Decimal(x):
-        return math.nextafter(z, math.inf)
+        return nextafter(z, math.inf)
     else:
         return z
 
@@ -675,7 +682,7 @@ def atanh_down(x: float) -> float:
     else:
         t = (d.exp() - 1) / (d.exp() + 1)
     if t > x:
-        return math.nextafter(y, -math.inf)
+        return nextafter(y, -math.inf)
     else:
         return y
 
@@ -687,7 +694,7 @@ def atanh_up(x: float) -> float:
     else:
         t = (d.exp() - 1) / (d.exp() + 1)
     if t > x:
-        return math.nextafter(y, math.inf)
+        return nextafter(y, math.inf)
     else:
         return y
 
@@ -950,9 +957,9 @@ def dist(p: Iterable[Union[Interval, RealLike]], q: Iterable[Union[Interval, Rea
             if len(partials) == 1 or partials[-2] > 0:
                 L = float_down(partials[-1])
             else:
-                L = math.nextafter(float_up(partials[-1]), -math.inf)
+                L = nextafter(float_up(partials[-1]), -math.inf)
             if L > 0.0 and math.isinf(L):
-                L = math.nextafter(L, 0.0)
+                L = nextafter(L, 0.0)
             partials = multi_add(*[
                 float_up(max(
                     Decimal(sub_intervals[i].minimum) - Decimal(sub_intervals[~i].maximum),
@@ -964,47 +971,133 @@ def dist(p: Iterable[Union[Interval, RealLike]], q: Iterable[Union[Interval, Rea
             if len(partials) == 1 or partials[-2] < 0:
                 U = float_up(partials[-1])
             else:
-                U = math.nextafter(float_down(partials[-1]), math.inf)
+                U = nextafter(float_down(partials[-1]), math.inf)
             if U < 0.0 and math.isinf(U):
-                U = math.nextafter(U, 0.0)
+                U = nextafter(U, 0.0)
             intervals.append((L, U))
     return sqrt(Interval(*intervals))
 
+def partials_add_exponent(
+    partials: List[Tuple[int, float]],
+    exponent: Optional[int] = None,
+    *mantissa: float,
+    others: Optional[List[Tuple[int, float]]] = None,
+) -> List[Tuple[int, float]]:
+    loop = ((exponent, mx) for mx in mantissa) if others is None else others
+    for me, mx in loop:
+        i = 0
+        for pe, px in partials:
+            if me < pe or abs(mx) < abs(px):
+                me, mx, pe, px = pe, px, me, mx
+            total = mx + math.ldexp(px, pe - me)
+            if math.isinf(total):
+                partials[:] = [(0, total)]
+                return partials
+            if total == mx:
+                px, normalization = math.frexp(px)
+                pe += normalization
+                partials[i] = (pe, px)
+                i += 1
+                mx, normalization = math.frexp(mx)
+                me += normalization
+                continue
+            error = math.ldexp(px, pe - me) - (total - mx)
+            print(error)
+            if error != 0.0:
+                error, normalization = math.frexp(error)
+                partials[i] = (me + normalization, error)
+                i += 1
+            mx, normalization = math.frexp(total)
+            me += normalization
+        partials[i:] = [(me, mx)]
+        if len(partials) > 1 and mx == 0.0:
+            del partials[-1]
+    return partials
+
+pae = partials_add_exponent
+
+def partials_down(partials: List[Tuple[int, float]]) -> float:
+    z = math.ldexp(*partials[-1][::-1])
+    z_mantissa, z_exponent = math.frexp(z)
+    if z_mantissa > math.ldexp(partials[-1][1], partials[-1][0] - z_exponent):
+        return nextafter(z, -math.inf)
+    elif z_mantissa < math.ldexp(partials[-1][1], partials[-1][0] - z_exponent):
+        return z
+    elif len(partials) == 1 or partials[-2][1] >= 0.0:
+        return z
+    else:
+        return nextafter(z, -math.inf)
+
+def partials_up(partials: List[Tuple[int, float]]) -> float:
+    z = math.ldexp(*partials[-1][::-1])
+    z_mantissa, z_exponent = math.frexp(z)
+    if z_mantissa < math.ldexp(partials[-1][1], partials[-1][0] - z_exponent):
+        return nextafter(z, math.inf)
+    elif z_mantissa > math.ldexp(partials[-1][1], partials[-1][0] - z_exponent):
+        return z
+    elif len(partials) == 1 or partials[-2][1] <= 0.0:
+        return z
+    else:
+        return nextafter(z, math.inf)
+
 def dot(x: Iterable[Union[Interval, RealLike]], y: Iterable[Union[Interval, RealLike]]) -> Interval:
-    with localcontext() as ctx:
-        ctx.prec = 50
-        intervals = [([], [])]
-        for xi, yi in zip(x, y):
-            if isinstance(xi, get_args(RealLike)):
-                if isinstance(xi, SupportsIndex):
-                    xi = operator.index(xi)
-                xi = Interval(float_split(xi))
-            elif not isinstance(xi, Interval):
-                raise TypeError(NOT_INTERVAL.format(repr(xi)))
-            if isinstance(yi, get_args(RealLike)):
-                if isinstance(yi, SupportsIndex):
-                    yi = operator.index(yi)
-                yi = Interval(float_split(yi))
-            elif not isinstance(yi, Interval):
-                raise TypeError(NOT_INTERVAL.format(repr(yi)))
-            lower = []
-            upper = []
-            for xii in xi[:0].sub_intervals:
-                for yii in yi[:0].sub_intervals:
-                    lower.append(Decimal(xii.maximum) * Decimal(yii.maximum))
-                    upper.append(Decimal(xii.minimum) * Decimal(yii.minimum))
-                for yii in yi[0:].sub_intervals:
-                    lower.append(Decimal(xii.minimum) * Decimal(yii.maximum))
-                    upper.append(Decimal(xii.maximum) * Decimal(yii.minimum))
-            for xii in xi[0:].sub_intervals:
-                for yii in yi[:0].sub_intervals:
-                    lower.append(Decimal(xii.maximum) * Decimal(yii.minimum))
-                    upper.append(Decimal(xii.minimum) * Decimal(yii.maximum))
-                for yii in yi[0:].sub_intervals:
-                    lower.append(Decimal(xii.minimum) * Decimal(yii.minimum))
-                    upper.append(Decimal(xii.maximum) * Decimal(yii.maximum))
-            intervals[:] = [(i[0] + L, i[1] + U) for i in intervals for L, U in zip(lower, upper)]
-        return Interval(*intervals)
+    lower = []
+    upper = []
+    lowers = [[(0, 0.0)]]
+    uppers = [[(0, 0.0)]]
+    for xi, yi in zip(x, y):
+        if isinstance(xi, get_args(RealLike)):
+            if isinstance(xi, SupportsIndex):
+                xi = operator.index(xi)
+            xi = Interval(float_split(xi))
+        elif not isinstance(xi, Interval):
+            raise TypeError(NOT_INTERVAL.format(repr(xi)))
+        if isinstance(yi, get_args(RealLike)):
+            if isinstance(yi, SupportsIndex):
+                yi = operator.index(yi)
+            yi = Interval(float_split(yi))
+        elif not isinstance(yi, Interval):
+            raise TypeError(NOT_INTERVAL.format(repr(yi)))
+        for xii in xi[:0].sub_intervals:
+            for yii in yi[:0].sub_intervals:
+                me, mx = mul_precise(xii.maximum, yii.maximum)
+                lower.append(pae([], me, *mx))
+                me, mx = mul_precise(xii.minimum, yii.minimum)
+                upper.append(pae([], me, *mx))
+            for yii in yi[0:].sub_intervals:
+                me, mx = mul_precise(xii.minimum, yii.maximum)
+                lower.append(pae([], me, *mx))
+                me, mx = mul_precise(xii.maximum, yii.minimum)
+                upper.append(pae([], me, *mx))
+        for xii in xi[0:].sub_intervals:
+            for yii in yi[:0].sub_intervals:
+                me, mx = mul_precise(xii.maximum, yii.minimum)
+                lower.append(pae([], me, *mx))
+                me, mx = mul_precise(xii.minimum, yii.maximum)
+                upper.append(pae([], me, *mx))
+            for yii in yi[0:].sub_intervals:
+                me, mx = mul_precise(xii.minimum, yii.minimum)
+                lower.append(pae([], me, *mx))
+                me, mx = mul_precise(xii.maximum, yii.maximum)
+                upper.append(pae([], me, *mx))
+        print(lower, upper)
+        lowers = [
+            pae(L1.copy(), others=L2)
+            for L1 in lower
+            for L2 in lowers
+        ]
+        lower.clear()
+        uppers = [
+            pae(U1.copy(), others=U2)
+            for U1 in upper
+            for U2 in uppers
+        ]
+        upper.clear()
+    print(lowers, uppers)
+    return Interval(*[
+        (partials_down(L), partials_up(U))
+        for L, U in zip(lowers, uppers)
+    ])
 
 def erf_small_precise(x: float) -> Decimal:
     assert abs(x) <= 1.5
@@ -1064,7 +1157,7 @@ def erf_down(x: float) -> float:
     elif math.isinf(x):
         return 1.0
     else:
-        return math.nextafter(1.0, 0.0)
+        return nextafter(1.0, 0.0)
 
 def erf_up(x: float) -> float:
     if abs(x) <= 1.5:
@@ -1078,7 +1171,7 @@ def erf_up(x: float) -> float:
     elif math.isinf(x):
         return -1.0
     else:
-        return math.nextafter(-1.0, 0.0)
+        return nextafter(-1.0, 0.0)
 
 def erfc(x: Union[Interval, RealLike]) -> Interval:
     if isinstance(x, get_args(RealLike)):
@@ -1105,7 +1198,7 @@ def erfc_down(x: float) -> float:
     elif math.isinf(x):
         return 2.0
     else:
-        return math.nextafter(2.0, 0.0)
+        return nextafter(2.0, 0.0)
 
 def erfc_up(x: float) -> float:
     if abs(x) <= 1.5:
@@ -1119,19 +1212,19 @@ def erfc_up(x: float) -> float:
     elif math.isinf(x):
         return 0.0
     else:
-        return math.nextafter(0.0, 1.0)
+        return nextafter(0.0, 1.0)
 
 def exp(x: Union[Interval, RealLike]) -> Interval:
     if isinstance(x, Decimal):
         try:
             return Interval(float_split(x.exp()))
         except decimal.Overflow:
-            return Interval((math.nextafter(math.inf, 0.0), math.inf))
+            return Interval((nextafter(math.inf, 0.0), math.inf))
     elif isinstance(x, SupportsIndex):
         try:
             return Interval(float_split(Decimal(operator.index(x)).exp()))
         except decimal.Overflow:
-            return Interval((math.nextafter(math.inf, 0.0), math.inf))
+            return Interval((nextafter(math.inf, 0.0), math.inf))
     elif isinstance(x, get_args(RealLike)):
         x = Interval(float_split(x))
     elif not isinstance(x, Interval):
@@ -1146,7 +1239,7 @@ def exp_down(x: float) -> float:
     try:
         return float_down(Decimal(x).exp())
     except decimal.Overflow:
-        return math.nextafter(math.inf, x)
+        return nextafter(math.inf, x)
 
 def exp_up(x: float) -> float:
     try:
@@ -1156,7 +1249,7 @@ def exp_up(x: float) -> float:
     if result != 0.0 or math.isinf(x):
         return result
     else:
-        return math.nextafter(0.0, 1.0)
+        return nextafter(0.0, 1.0)
 
 def expm1_precise(x: float) -> Decimal:
     with localcontext() as ctx:
@@ -1178,12 +1271,12 @@ def expm1(x: Union[Interval, RealLike]) -> Interval:
         try:
             return Interval(float_split(x.exp() - 1))
         except decimal.Overflow:
-            return Interval((math.nextafter(math.inf, 0.0), math.inf))
+            return Interval((nextafter(math.inf, 0.0), math.inf))
     elif isinstance(x, SupportsIndex):
         try:
             return Interval(float_split(Decimal(operator.index(x)).exp() - 1))
         except decimal.Overflow:
-            return Interval((math.nextafter(math.inf, 0.0), math.inf))
+            return Interval((nextafter(math.inf, 0.0), math.inf))
     elif isinstance(x, get_args(RealLike)):
         x = Interval(float_split(x))
     else:
@@ -1198,7 +1291,7 @@ def expm1_down(x: float) -> float:
     try:
         return float_down(expm1_precise(x))
     except decimal.Overflow:
-        return math.nextafter(math.inf, x)
+        return nextafter(math.inf, x)
 
 def expm1_up(x: float) -> float:
     try:
@@ -1208,7 +1301,7 @@ def expm1_up(x: float) -> float:
     if result != -1.0 or math.isinf(x):
         return result
     else:
-        return math.nextafter(-1.0, 0.0)
+        return nextafter(-1.0, 0.0)
 
 def fsum(intervals: Iterable[Union[Interval, RealLike]]) -> Interval:
     temp = []
@@ -1239,9 +1332,9 @@ def fsum(intervals: Iterable[Union[Interval, RealLike]]) -> Interval:
             if len(partials) == 1 or partials[-2] > 0:
                 L = float_down(partials[-1])
             else:
-                L = math.nextafter(float_up(partials[-1]), -math.inf)
+                L = nextafter(float_up(partials[-1]), -math.inf)
             if L > 0.0 and math.isinf(L):
-                L = math.nextafter(L, 0.0)
+                L = nextafter(L, 0.0)
             partials = multi_add(*[
                 sub_interval.maximum
                 for sub_interval in sub_intervals
@@ -1249,9 +1342,9 @@ def fsum(intervals: Iterable[Union[Interval, RealLike]]) -> Interval:
             if len(partials) == 1 or partials[-2] < 0:
                 U = float_up(partials[-1])
             else:
-                U = math.nextafter(float_down(partials[-1]), math.inf)
+                U = nextafter(float_down(partials[-1]), math.inf)
             if U < 0.0 and math.isinf(U):
-                U = math.nextafter(U, 0.0)
+                U = nextafter(U, 0.0)
             results.append((L, U))
     return Interval(*results)
 
@@ -1262,9 +1355,9 @@ LGAMMA_POS_MIN_Y = Decimal(
     "-0.1214862905358496080955145571"
 )
 
-DIGAMMA_ROOTS_CACHE: dict[int, tuple[float, float]] = {}
+DIGAMMA_ROOTS_CACHE: Dict[int, Tuple[float, float]] = {}
 
-def digamma_root(n: int) -> tuple[float, float]:
+def digamma_root(n: int) -> Tuple[float, float]:
     if n not in DIGAMMA_ROOTS_CACHE:
         L = 0.0
         U = 0.5
@@ -1284,7 +1377,7 @@ def digamma_root(n: int) -> tuple[float, float]:
         DIGAMMA_ROOTS_CACHE[n] = (L, U)
     return DIGAMMA_ROOTS_CACHE[n]
 
-LGAMMA_MINS_CACHE: dict[int, Decimal] = {}
+LGAMMA_MINS_CACHE: Dict[int, Decimal] = {}
 
 def lgamma_min(n: int) -> float:
     if n not in LGAMMA_MINS_CACHE:
@@ -1385,7 +1478,7 @@ def lgamma_down(x: float) -> float:
         return math.inf
     y = float_down(lgamma_precise(x))
     if 0 < x < math.inf and math.isinf(y):
-        return math.nextafter(y, 0.0)
+        return nextafter(y, 0.0)
     else:
         return y
 
@@ -1418,7 +1511,7 @@ def gamma(x: Union[Interval, RealLike]) -> Interval:
         if x <= 0:
             return Interval((-math.inf, -math.inf), (math.inf, math.inf))
         elif x >= 179:
-            return Interval((math.nextafter(math.inf, 0.0), math.inf))
+            return Interval((nextafter(math.inf, 0.0), math.inf))
         else:
             return Interval(float_split(gamma_precise(float(x))))
     elif isinstance(x, SupportsRichFloat):
@@ -1528,7 +1621,7 @@ def gamma(x: Union[Interval, RealLike]) -> Interval:
 def gamma_down(x: float) -> float:
     y = float_down(gamma_precise(x))
     if 0 < x < math.inf and math.isinf(y):
-        return math.nextafter(y, 0.0)
+        return nextafter(y, 0.0)
     else:
         return y
 
@@ -1914,21 +2007,21 @@ def sqrt_down(x: float) -> float:
     y = math.sqrt(x)
     partials = mul_precise(y, y)
     if partials[-1] > x:
-        return math.nextafter(y, 0.0)
+        return nextafter(y, 0.0)
     elif partials[-1] < x or len(partials) == 1 or partials[-2] < 0.0:
         return y
     else:
-        return math.nextafter(y, 0.0)
+        return nextafter(y, 0.0)
 
 def sqrt_up(x: float) -> float:
     y = math.sqrt(x)
     partials = mul_precise(y, y)
     if partials[-1] < x:
-        return math.nextafter(y, math.inf)
+        return nextafter(y, math.inf)
     elif partials[-1] > x or len(partials) == 1 or partials[-2] > 0.0:
         return y
     else:
-        return math.nextafter(y, math.inf)
+        return nextafter(y, math.inf)
 
 def tan(x: Union[Interval, RealLike]) -> Interval:
     if isinstance(x, (Decimal, SupportsIndex)):
